@@ -1,8 +1,10 @@
 import copy
+import functools
+import re
 from enum import Enum
 from typing import Any, Dict, Tuple, Type
 
-import pytest
+from pytest import mark, raises
 
 from omegaconf import (
     AnyNode,
@@ -12,17 +14,22 @@ from omegaconf import (
     FloatNode,
     IntegerNode,
     ListConfig,
+    Node,
     OmegaConf,
     StringNode,
     ValueNode,
 )
-from omegaconf.errors import ValidationError
-
-from . import Color, IllegalType, User
+from omegaconf.errors import (
+    InterpolationToMissingValueError,
+    UnsupportedValueType,
+    ValidationError,
+)
+from omegaconf.nodes import InterpolationResultNode
+from tests import Color, IllegalType, User
 
 
 # testing valid conversions
-@pytest.mark.parametrize(  # type: ignore
+@mark.parametrize(
     "type_,input_,output_",
     [
         # string
@@ -81,7 +88,7 @@ def test_valid_inputs(type_: type, input_: Any, output_: Any) -> None:
 
 
 # testing invalid conversions
-@pytest.mark.parametrize(  # type: ignore
+@mark.parametrize(
     "type_,input_",
     [
         (IntegerNode, "abc"),
@@ -117,13 +124,13 @@ def test_valid_inputs(type_: type, input_: Any, output_: Any) -> None:
 def test_invalid_inputs(type_: type, input_: Any) -> None:
     empty_node = type_()
 
-    with pytest.raises(ValidationError):
+    with raises(ValidationError):
         empty_node._set_value(input_)
-    with pytest.raises(ValidationError):
+    with raises(ValidationError):
         type_(input_)
 
 
-@pytest.mark.parametrize(  # type: ignore
+@mark.parametrize(
     "input_, expected_type",
     [
         ({}, DictConfig),
@@ -192,14 +199,14 @@ def test_list_integer_rejects_string() -> None:
     assert isinstance(c, ListConfig)
     c.append(IntegerNode(10))
     assert c.get(0) == 10
-    with pytest.raises(ValidationError):
+    with raises(ValidationError):
         c[0] = "string"
     assert c[0] == 10
     assert type(c._get_node(0)) == IntegerNode
 
 
 # Test merge raises validation error
-@pytest.mark.parametrize(  # type: ignore
+@mark.parametrize(
     "c1, c2",
     [
         (dict(a=IntegerNode(10)), dict(a="str")),
@@ -211,14 +218,14 @@ def test_list_integer_rejects_string() -> None:
 def test_merge_validation_error(c1: Dict[str, Any], c2: Dict[str, Any]) -> None:
     conf1 = OmegaConf.create(c1)
     conf2 = OmegaConf.create(c2)
-    with pytest.raises(ValidationError):
+    with raises(ValidationError):
         OmegaConf.merge(conf1, conf2)
     # make sure that conf1 and conf2 were not modified
     assert conf1 == OmegaConf.create(c1)
     assert conf2 == OmegaConf.create(c2)
 
 
-@pytest.mark.parametrize(  # type: ignore
+@mark.parametrize(
     "type_,valid_value, invalid_value",
     [
         (IntegerNode, 1, "invalid"),
@@ -247,7 +254,7 @@ def test_accepts_mandatory_missing(
     assert conf.foo == valid_value
 
     if invalid_value is not None:
-        with pytest.raises(ValidationError):
+        with raises(ValidationError):
             conf.foo = invalid_value
 
 
@@ -261,10 +268,10 @@ class Enum2(Enum):
     NOT_BAR = 2
 
 
-@pytest.mark.parametrize(  # type: ignore
+@mark.parametrize(
     "type_", [BooleanNode, EnumNode, FloatNode, IntegerNode, StringNode, AnyNode]
 )
-@pytest.mark.parametrize(  # type: ignore
+@mark.parametrize(
     "values, success_map",
     [
         (
@@ -328,11 +335,11 @@ def test_legal_assignment(
             node = type_(value)
             assert node._value() == expected
         else:
-            with pytest.raises(ValidationError):
+            with raises(ValidationError):
                 type_(value)
 
 
-@pytest.mark.parametrize(  # type: ignore
+@mark.parametrize(
     "node,value",
     [
         (IntegerNode(), "foo"),
@@ -342,14 +349,14 @@ def test_legal_assignment(
     ],
 )
 def test_illegal_assignment(node: ValueNode, value: Any) -> None:
-    with pytest.raises(ValidationError):
+    with raises(ValidationError):
         node._set_value(value)
 
 
-@pytest.mark.parametrize(  # type: ignore
+@mark.parametrize(
     "node_type", [BooleanNode, EnumNode, FloatNode, IntegerNode, StringNode, AnyNode]
 )
-@pytest.mark.parametrize(  # type: ignore
+@mark.parametrize(
     "enum_type, values, success_map",
     [
         (
@@ -362,7 +369,7 @@ def test_illegal_assignment(node: ValueNode, value: Any) -> None:
 def test_legal_assignment_enum(
     node_type: Type[EnumNode],
     enum_type: Type[Enum],
-    values: Tuple[Any],
+    values: Tuple[Any, ...],
     success_map: Dict[Any, Any],
 ) -> None:
     assert isinstance(values, (list, tuple))
@@ -376,7 +383,7 @@ def test_legal_assignment_enum(
             node._set_value(value)
             assert node._value() == expected
         else:
-            with pytest.raises(ValidationError):
+            with raises(ValidationError):
                 node_type(enum_type)
 
 
@@ -384,8 +391,8 @@ class DummyEnum(Enum):
     FOO = 1
 
 
-@pytest.mark.parametrize("is_optional", [True, False])  # type: ignore
-@pytest.mark.parametrize(  # type: ignore
+@mark.parametrize("is_optional", [True, False])
+@mark.parametrize(
     "ref_type, type_, value, expected_type",
     [
         (Any, Any, 10, AnyNode),
@@ -433,13 +440,13 @@ def test_node_wrap_illegal_type() -> None:
 
     from omegaconf.omegaconf import _node_wrap
 
-    with pytest.raises(ValidationError):
+    with raises(ValidationError):
         _node_wrap(
             type_=UserClass, value=UserClass(), is_optional=False, parent=None, key=None
         )
 
 
-@pytest.mark.parametrize(  # type: ignore
+@mark.parametrize(
     "obj",
     [
         StringNode(),
@@ -462,7 +469,7 @@ def test_deepcopy(obj: Any) -> None:
         assert obj.__dict__[k] == cp.__dict__[k]
 
 
-@pytest.mark.parametrize(  # type: ignore
+@mark.parametrize(
     "node, value, expected",
     [
         (StringNode(), None, True),
@@ -483,12 +490,7 @@ def test_deepcopy(obj: Any) -> None:
         (BooleanNode(True), None, False),
         (BooleanNode(True), False, False),
         (BooleanNode(False), False, True),
-        (AnyNode(value=1, is_optional=True), AnyNode(value=1, is_optional=True), True),
-        (
-            AnyNode(value=1, is_optional=True),
-            AnyNode(value=1, is_optional=False),
-            True,
-        ),
+        (AnyNode(value=1), AnyNode(value=1), True),
         (EnumNode(enum_type=Enum1), Enum1.BAR, False),
         (EnumNode(enum_type=Enum1), EnumNode(Enum1), True),
         (EnumNode(enum_type=Enum1), "nope", False),
@@ -498,6 +500,18 @@ def test_deepcopy(obj: Any) -> None:
             True,
         ),
         (EnumNode(enum_type=Enum1, value=Enum1.BAR), Enum1.BAR, True),
+        (InterpolationResultNode("foo"), "foo", True),
+        (InterpolationResultNode("${foo}"), "${foo}", True),
+        (InterpolationResultNode("${foo"), "${foo", True),
+        (InterpolationResultNode(None), None, True),
+        (InterpolationResultNode(1), 1, True),
+        (InterpolationResultNode(1.0), 1.0, True),
+        (InterpolationResultNode(True), True, True),
+        (InterpolationResultNode(Color.RED), Color.RED, True),
+        (InterpolationResultNode({"a": 0, "b": 1}), {"a": 0, "b": 1}, True),
+        (InterpolationResultNode([0, None, True]), [0, None, True], True),
+        (InterpolationResultNode("foo"), 100, False),
+        (InterpolationResultNode(100), "foo", False),
     ],
 )
 def test_eq(node: ValueNode, value: Any, expected: Any) -> None:
@@ -505,10 +519,13 @@ def test_eq(node: ValueNode, value: Any, expected: Any) -> None:
     assert (node != value) != expected
     assert (value == node) == expected
     assert (value != node) != expected
-    assert (node.__hash__() == value.__hash__()) == expected
+
+    # Check hash except for unhashable types (dict/list).
+    if not isinstance(value, (dict, list)):
+        assert (node.__hash__() == value.__hash__()) == expected
 
 
-@pytest.mark.parametrize("value", [1, 3.14, True, None, Enum1.FOO])  # type: ignore
+@mark.parametrize("value", [1, 3.14, True, None, Enum1.FOO])
 def test_set_anynode_with_primitive_type(value: Any) -> None:
     cfg = OmegaConf.create({"a": 5})
     a_before = cfg._get_node("a")
@@ -518,7 +535,7 @@ def test_set_anynode_with_primitive_type(value: Any) -> None:
     assert cfg.a == value
 
 
-@pytest.mark.parametrize(  # type: ignore
+@mark.parametrize(
     "value, container_type",
     [
         (ListConfig(content=[1, 2]), ListConfig),
@@ -539,7 +556,7 @@ def test_set_anynode_with_container(value: Any, container_type: Any) -> None:
 
 def test_set_anynode_with_illegal_type() -> None:
     cfg = OmegaConf.create({"a": 5})
-    with pytest.raises(ValidationError):
+    with raises(ValidationError):
         cfg.a = IllegalType()
 
 
@@ -548,5 +565,104 @@ def test_set_valuenode() -> None:
     a_before = cfg._get_node("age")
     cfg.age = 12
     assert id(cfg._get_node("age")) == id(a_before)
-    with pytest.raises(ValidationError):
+    with raises(ValidationError):
         cfg.age = []
+
+
+def test_allow_objects() -> None:
+    c = OmegaConf.create({"foo": AnyNode()})
+    with raises(UnsupportedValueType):
+        c.foo = IllegalType()
+    c._set_flag("allow_objects", True)
+    iv = IllegalType()
+    c.foo = iv
+    assert c.foo == iv
+
+
+def test_dereference_missing() -> None:
+    cfg = OmegaConf.create({"x": "???"})
+    x_node = cfg._get_node("x")
+    assert isinstance(x_node, Node)
+    assert x_node._dereference_node() is x_node
+
+
+@mark.parametrize(
+    "make_func",
+    [
+        StringNode,
+        IntegerNode,
+        FloatNode,
+        BooleanNode,
+        lambda val, is_optional: EnumNode(
+            enum_type=Color, value=val, is_optional=is_optional
+        ),
+    ],
+)
+def test_validate_and_convert_none(make_func: Any) -> None:
+    node = make_func("???", is_optional=False)
+    with raises(
+        ValidationError, match=re.escape("Non optional field cannot be assigned None")
+    ):
+        node.validate_and_convert(None)
+
+
+def test_dereference_interpolation_to_missing() -> None:
+    cfg = OmegaConf.create({"x": "${y}", "y": "???"})
+    x_node = cfg._get_node("x")
+    assert isinstance(x_node, Node)
+    assert x_node._maybe_dereference_node() is None
+    with raises(InterpolationToMissingValueError):
+        cfg.x
+
+
+@mark.parametrize(
+    "flags",
+    [
+        {},
+        {"flag": True},
+        {"flag": False},
+        {"flag1": True, "flag2": False},
+    ],
+)
+@mark.parametrize(
+    "type_",
+    [
+        AnyNode,
+        BooleanNode,
+        functools.partial(EnumNode, enum_type=Color),
+        FloatNode,
+        IntegerNode,
+        InterpolationResultNode,
+        StringNode,
+    ],
+)
+def test_set_flags_in_init(type_: Any, flags: Dict[str, bool]) -> None:
+    node = type_(value=None, flags=flags)
+    for f, v in flags.items():
+        assert node._get_flag(f) is v
+
+
+@mark.parametrize(
+    "flags",
+    [
+        None,
+        {"flag": True},
+        {"flag": False},
+        {"readonly": True},
+        {"readonly": False},
+        {"flag1": True, "flag2": False, "readonly": False},
+        {"flag1": False, "flag2": True, "readonly": True},
+    ],
+)
+def test_interpolation_result_readonly(flags: Any) -> None:
+    readonly = None if flags is None else flags.get("readonly")
+    expected = [] if flags is None else list(flags.items())
+    node = InterpolationResultNode("foo", flags=flags)
+
+    # Check that flags are set to their desired value.
+    for k, v in expected:
+        assert node._get_node_flag(k) is v
+
+    # If no value was provided for the "readonly" flag, it should be set.
+    if readonly is None:
+        assert node._get_node_flag("readonly")
